/**
 * DataList
 *
 * Provides a reactive View of a DataStore
 * Requires a CSS class named "hidden" to toggle "display:none" of list items
 *
 * @class datalist
 * @namespace abaaso
 */
var datalist = {
	/**
	 * Creates an instance of datalist
	 *
	 * Events: beforeDataList  Fires before target receives the DataList
	 *         afterDataList   Fires after DataList is setup in target
	 *           
	 * @method factory
	 * @param  {Object} target   Element to receive the DataList
	 * @param  {Object} store    Data store to feed the DataList
	 * @param  {Mixed}  template Record field, template ($.tpl), or String, e.g. "<p>this is a {{field}} sample.</p>", fields are marked with {{ }}
	 * @param  {Object} options  Optional parameters to set on the DataList
	 * @return {Object}          DataList instance
	 */
	factory : function (target, store, template, options) {
		var ref = [store],
		    element, instance;

		if (!(target instanceof Element) || typeof store !== "object" || !regex.string_object.test(typeof template)) throw Error(label.error.invalidArguments);

		element = target.fire("beforeDataList")
		                .create("ul", {"class": "list", id: store.parentNode.id + "-datalist"});

		// Creating instance
		instance = new DataList(element, ref[0], template);
		if (options instanceof Object) utility.merge(instance, options);
		instance.store.datalists.push(instance);

		// Rendering if not tied to an API or data is ready
		if (instance.store.uri === null || instance.store.loaded) instance.display();

		target.fire("afterDataList", element);
		return instance;
	},

	// Inherited by DataLists
	methods : {
		/**
		 * Delete sync handler
		 * 
		 * @method del
		 * @param  {Object} rec Record
		 * @return {Undefined}  undefined
		 */
		del : function (rec) {
			if (typeof this.pageIndex === "number" && typeof this.pageSize === "number") this.refresh();
			else {
				this.element.fire("beforeDataListRefresh");
				this.element.find("> li[data-key='" + rec.key + "']").destroy();
				this.element.fire("afterDataListRefresh");
			}
		},

		/**
		 * Displays the data list (unpause)
		 * 
		 * @method display
		 * @return {Undefined} undefined
		 */
		display : function () {
			this.ready = true;
			this.refresh(true);
		},

		/**
		 * Changes the page index of the DataList
		 * 
		 * @method page
		 * @return {Object}  DataList instance
		 */
		page : function (arg) {
			if (isNaN(arg)) throw Error(label.error.invalidArguments);

			this.pageIndex = arg;
			this.refresh();
			return this;
		},

		/**
		 * Adds pagination Elements to the View
		 * 
		 * @method pages
		 * @return {Object}  DataList instance
		 */
		pages : function () {
			var obj   = this.element,
			    list  = $("#" + obj.id + "-pages-top, #" + obj.id + "-pages-bottom"),
			    page  = this.pageIndex,
			    pos   = this.pagination,
			    range = this.pageRange,
			    mid   = number.round(number.half(range), "down"),
			    start = page - mid,
			    end   = page + mid,
			    self  = this,
			    total = datalist.pages.call(this),
			    i     = 0,
			    diff, li, anchor;

			if (!regex.top_bottom.test(pos)) throw Error(label.error.invalidArguments);

			// Removing the existing controls
			array.each(list, function (i) {
				if (i !== undefined) i.destroy();
			});
			
			// Halting because there's 1 page, or nothing
			if (this.total === 0 || total === 1) return this;

			// Getting the range to display
			if (start < 1) {
				diff  = start.diff(1);
				start = start + diff;
				end   = end   + diff;
			}
			if (end > total) {
				end   = total;
				start = (end - range) + 1;
				if (start < 1) start = 1;
			}

			array.each(pos.explode(), function (i) {
				var current = false,
				    more    = page > 1,
				    next    = (page + 1) <= total,
				    last    = !(page < total);

				// Setting up the list
				list = obj[i === "bottom" ? "after" : "before"]("ul", {"class": "list pages " + i, id: obj.id + "-pages-" + i});

				// First page
				list.create("li").create(more ? "a" : "span", {"class": "first page", "data-page": 1}).html("&lt;&lt;");

				// Previous page
				list.create("li").create(more ? "a" : "span", {"class": "prev page", "data-page": (page - 1)}).html("&lt;");

				// Rendering the page range
				for (i = start; i <= end; i++) {
					current = (i === page);
					list.create("li").create(current ? "span" : "a", {"class": current ? "current page" : "page", "data-page": i}).html(i);
				}

				// Next page
				list.create("li").create(next ? "a" : "span", {"class": "next page", "data-page": next ? (page + 1) : null}).html("&gt;");

				// Last page
				list.create("li").create(last ? "span" : "a", {"class": "last page", "data-page": last ? null : total}).html("&gt;&gt;");

				// Scroll to top the top
				list.find("a.page").on("click", function (e) {
					utility.stop(e);
					self.page(this.data("page"));
					window.scrollTo(0, 0);
				}, "pagination");
			});

			return this;
		},

		/**
		 * Refreshes element
		 * 
		 * Events: beforeDataListRefresh  Fires from the element containing the DataList
		 *         afterDataListRefresh   Fires from the element containing the DataList
		 * 
		 * @method refresh
		 * @param {Boolean} redraw [Optional] Boolean to force clearing the DataList (default), false toggles "hidden" class of items
		 * @return {Object}        DataList instance
		 */
		refresh : function (redraw) {
			redraw       = (redraw !== false);
			var element  = this.element,
			    template = (typeof this.template === "object"),
			    key      = (!template && String(this.template).replace(/\{\{|\}\}/g, "") === this.store.key),
			    consumed = [],
			    items    = [],
			    self     = this,
			    callback = (typeof this.callback === "function"),
			    reg      = new RegExp(),
			    registry = [], // keeps track of records in the list (for filtering)
			    limit    = [],
			    fn, obj, ceiling;

			this.element.fire("beforeDataListRefresh");
			this.refreshing = true;

			// Creating templates for the html rep
			if (!template) fn = function (i) {
				var html  = self.template,
				    items = array.unique(html.match(/\{\{[\w\.]+\}\}/g));

				// Replacing record key
				html = html.replace("{{" + self.store.key + "}}", i.key);
				
				// Replacing dot notation properties
				array.each(items, function (attr) {
					var key   = attr.replace(/\{\{|\}\}/g, ""),
					    value = utility.walk(i.data, key);

					reg.compile(attr, "g");
					html = html.replace(reg, value);
				});

				// Filling in placeholder value
				html = html.replace(/\{\{.*\}\}/g, self.placeholder);

				return {li: html};
			}
			else fn = function (i) {
				var obj   = json.encode(self.template),
				    items = array.unique(obj.match(/\{\{[\w\.]+\}\}/g));

				// Replacing record key
				obj = obj.replace("{{" + self.store.key + "}}", i.key);
				
				// Replacing dot notation properties
				array.each(items, function (attr) {
					var key   = attr.replace(/\{\{|\}\}/g, ""),
					    value = utility.walk(i.data, key);

					reg.compile(attr, "g");

					// Stripping first and last " to concat to valid JSON
					obj = obj.replace(reg, json.encode(value).replace(/(^")|("$)/g, ""));
				});

				// Filling in placeholder value
				obj = json.decode(obj.replace(/\{\{.*\}\}/g, self.placeholder));

				return {li: obj};
			};

			// Consuming records based on sort
			if (this.where === null) consumed = this.order.isEmpty() ? this.store.get() : this.store.sort(this.order, false, this.sensitivity);
			else consumed = this.order.isEmpty() ? this.store.select(this.where) : this.store.sort(this.order, false, this.sensitivity, this.where);

			// Processing (filtering) records & generating templates
			array.each(consumed, function (i) {
				if (self.filter === null || !(self.filter instanceof Object)) items.push({key: i.key, template: fn(i)});
				else {
					utility.iterate(self.filter, function (v, k) {
						if (registry.index(i.key) > -1) return;

						var x     = 0,
						    regex = new RegExp(),
						    nth;

						v   = string.explode(v);
						nth = v.length;

						for (x = 0; x < nth; x++) {
							regex.compile(v[x], "i");
							if ((k === self.store.key && regex.test(i.key)) || (i.data[k] !== undefined && regex.test(i.data[k]))) {
								registry.push(i.key);
								items.push({key: i.key, template: fn(i)});
								return;
							}
						}
					});
				}
			});

			// Total count of items in the list
			this.total = items.length;

			// Only do these ops if there's something to render
			if (this.total > 0) {
				// Pagination (supports filtering)
				if (typeof this.pageIndex === "number" && typeof this.pageSize === "number") {
					ceiling = datalist.pages.call(this);
					// Passed the end, so putting you on the end
					if (ceiling > 0 && this.pageIndex > ceiling) return this.page(ceiling);
					// Paginating the items
					else {
						limit = datalist.range.call(this);
						items = items.limit(limit[0], limit[1]);
					}
				}

				// Preparing the target element
				if (redraw) {
					element.clear();
					array.each(items, function (i) {
						var obj = element.tpl(i.template);
						obj.data("key", i.key);
						if (callback) self.callback(obj);
					});
				}
				else {
					element.find("> li").addClass("hidden");
					array.each(items, function (i) {
						element.find("> li[data-key='" + i.key + "']").removeClass("hidden");
					});
				}

				// Rendering pagination elements
				if (regex.top_bottom.test(this.pagination) && typeof this.pageIndex === "number" && typeof this.pageSize === "number") this.pages();
				else {
					$("#" + this.element.id + "-pages-top, #" + this.element.id + "-pages-bottom");
					if (obj !== undefined) obj.destroy();
				}
			}

			this.refreshing = false;
			this.element.fire("afterDataListRefresh", element);
			return this;
		},

		/**
		 * Set sync handler
		 *
		 * @method set
		 * @return {Undefined} undefined
		 */
		set : function () {
			if (!this.refreshing) this.refresh();
		},

		/**
		 * Sorts data list & refreshes element
		 * 
		 * Events: beforeDataListSort     Fires before the DataList sorts
		 *         afterDataListSort      Fires after the DataList is sorted
		 *         beforeDataListRefresh  Fires before the DataList refreshes
		 *         afterDataListRefresh   Fires after the DataList refreshes
		 * 
		 * @method sort
		 * @param  {String} order       SQL "order by" statement
		 * @param  {String} sensitivity [Optional] Defaults to "ci" ("ci" = insensitive, "cs" = sensitive, "ms" = mixed sensitive)
		 * @return {Object}              DataList instance
		 */
		sort : function (order, sensitivity) {
			if (typeof order !== "string") throw Error(label.error.invalidArguments);
			this.element.fire("beforeDataListSort");
			this.order       = order;
			this.sensitivity = sensitivity || "ci";
			this.refresh();
			this.element.fire("afterDataListSort");
			return this;
		},

		/**
		 * Tears down references to the DataList
		 * 
		 * @method teardown
		 * @param  {Boolean} destroy [Optional] `true` will remove the DataList from the DOM
		 * @return {Object}  DataList instance
		 */
		teardown : function (destroy) {
			destroy  = (destroy === true);
			var self = this,
			    id   = this.element.id;

			observer.remove(id);

			array.each($("#" + id + "-pages-top, #" + id + "-pages-bottom"), function (i) {
				observer.remove(i);
			});

			array.each(this.store.datalists, function (i, idx) {
				if (i.id === self.id) {
					this.remove(idx);
					return false;
				}
			});

			if (destroy) {
				element.destroy(this.element);
				this.element = null;
			}

			return this;
		}
	},

	/**
	 * Calculates the total pages
	 * 
	 * @method pages
	 * @return {Number} Total pages
	 */
	pages : function () {
		if (isNaN(this.pageSize)) throw Error(label.error.invalidArguments);
		return number.round(this.total / this.pageSize, "up");
	},

	/**
	 * Calculates the page size as an Array of start & finish
	 * 
	 * @method range
	 * @return {Array}  Array of start & end numbers
	 */
	range : function () {
		var start = (this.pageIndex * this.pageSize) - this.pageSize,
		    end   = this.pageSize;

		return [start, end];
	}
};

/**
 * DataList factory
 *
 * @class DataList
 * @namespace abaaso
 * @param  {Object} element  DataList element
 * @param  {Object} store    Data store to feed the DataList
 * @param  {Mixed}  template Record field, template ($.tpl), or String, e.g. "<p>this is a {{field}} sample.</p>", fields are marked with {{ }}
 * @return {Object}          Instance of DataList
 */
function DataList (element, store, template) {
	this.callback    = null;
	this.element     = element;
	this.filter      = null;
	this.id          = utility.genId();
	this.pageIndex   = 1;
	this.pageSize    = null;
	this.pageRange   = 5;
	this.pagination  = "bottom"; // "top" or "bottom|top" are also valid
	this.placeholder = "";
	this.order       = "";
	this.ready       = false;
	this.refreshing  = false;
	this.template    = template;
	this.total       = 0;
	this.sensitivity = "ci";
	this.store       = store;
	this.where       = null;
};

// Setting prototype & constructor loop
DataList.prototype = datalist.methods;
DataList.prototype.constructor = DataList;